// Copyright 2018 The Simons Foundation, Inc. - All Rights Reserved.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//    http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

#ifndef NETKET_MACHINE_HPP
#define NETKET_MACHINE_HPP

#include <fstream>
#include <memory>

#include "Graph/graph.hpp"
#include "Operator/operator.hpp"
#include "abstract_machine.hpp"
#include "ffnn.hpp"
#include "jastrow.hpp"
#include "jastrow_symm.hpp"
#include "mps_periodic.hpp"
#include "rbm_multival.hpp"
#include "rbm_spin.hpp"
#include "rbm_spin_phase.hpp"
#include "rbm_spin_real.hpp"
#include "rbm_spin_symm.hpp"

#include "DensityMatrices/density_matrix_machine.hpp"

#endif
